const winston = require('winston')

let config = {
  bucketName: null,
  storageFile: null,
  logFile: null,
  resourceNameSingular: null,
  resourceNamePlural: null,
  sortNameSingular: null,
  sortNamePlural: null,
  paginate: null,
  raiseEvents: false,
  configured: false
}

let store = {}

const make = (configuration) => {
  if (!Array.isArray(configuration)) throw new Error('Type Error: Parameter must be an array')
  configuration.forEach((option, index) => {
    // option[0] is key
    // option[1] is value
    // option[2] is config store key that will be given a proxy (e.g. resourceNameSingular)

    // Verifying the integrity of the configuration option structure
    if (!Array.isArray(option)) throw new Error(`Type Error: Parameter option ${index} must be an array`)
    if (option.length < 2) throw new Error(`Parameter option ${index} is missing a value. Should contain at least 2 elements`)

    // Set key and value in config store
    config[option[0]] = option[1]

    // Creating the getters and setters for the value
    store[option[0]] = () => config[option[0]]
    store[option[0]].set = (val) => config[option[0]] = val

    // sets config store key to a getter that returns the value (i.e. option[1]) of key (i.e. option[0])
    if (option.length === 3) {
      config[option[2]] = store[option[0]]()
    }
  })
  if (!config.configured) {
    // Checking for required configurations
    if (!config.storageFile ||
        !config.bucketName ||
        !config.logFile ||
        !config.resourceNamePlural ||
        !config.resourceNameSingular
    ) throw new Error('Missing required configuration option(s)')

    winston.info('Initialized bucket configurations')

    // Checking optional sortName configuration for correct implementation
    if (!config.sortNamePlural && !config.sortNameSingular) {
      winston.warn('No sortNames specified')
    } else if (config.sortNameSingular ? !config.sortNamePlural : config.sortNamePlural) {
      throw new Error('Missing one of two sortName properties in the controller configurations')
    }

    // Checking for optional paginate configuration
    if (!config.paginate) winston.warn('No pagination options specified')

    // Checking for optional raiseEvents configuration
    if (!config.raiseEvents) winston.warn('This API is currently configured to not raise events')

    // Setting configured storage
    config.configured = true
  }
}
const initialize = (configuration) => {
  winston.warn('This method is deprecated and will be removed in the future; init to be replaced by make.')
  if (config.configured) throw new Error('Controller configurations already set. Please refer to the make function.')
  const props = Object.keys(configuration)
  if (!props.includes('bucketName') ||
      !props.includes('storageFile') ||
      !props.includes('logFile') ||
      !props.includes('resourceNameSingular') ||
      !props.includes('resourceNamePlural')
  ) throw new Error('Missing required configuration option(s). Please check your identity-code-api-controllers setup.')

  winston.info('Initializing bucket configurations')

  if (props.includes('sortNameSingular') && props.includes('sortNamePlural')) {
    config.sortNameSingular = configuration.sortNameSingular || null
    config.sortNamePlural = configuration.sortNamePlural || null
  } else if (props.includes('sortNameSingular') ? !props.includes('sortNamePlural') : props.includes('sortNamePlural')) {
    throw new Error('Missing one of two sortName properties in the controller configurations')
  } else winston.info('No sortNames specified')

  if (props.includes('paginate')) {
    config.paginate = {
      default_start: (configuration.paginate.default_start ? configuration.paginate.default_start : 1),
      default_size: (configuration.paginate.default_size ? configuration.paginate.default_size : 50),
      max_page_size: (configuration.paginate.max_page_size ? configuration.paginate.max_page_size : 100)
    }
  } else winston.info('No pagination option(s) specified')

  if (props.includes('raiseEvents')) {
    config.raiseEvents = configuration.raiseEvents
  } else winston.info('Event raising is not configured')

  config.bucketName = configuration.bucketName
  config.storageFile = configuration.storageFile
  config.logFile = configuration.logFile
  config.resourceNameSingular = configuration.resourceNameSingular
  config.resourceNamePlural = configuration.resourceNamePlural
  config.configured = true
}
const configured = () => config.configured
const paginate = () => !!config.paginate
const raiseEvents = () => config.raiseEvents
const retrieve = {
  bucketName: () => config.bucketName,
  storageFile: () => config.storageFile,
  logFile: () => config.logFile,
  resourceNameSingular: () => config.resourceNameSingular,
  resourceNamePlural: () => config.resourceNamePlural,
  sortNameSingular: () => config.sortNameSingular,
  sortNamePlural: () => config.sortNamePlural,
  pageDefaultStart: () => config.paginate.default_start,
  pageDefaultSize: () => config.paginate.default_size,
  pageMaxSize: () => config.paginate.max_page_size,
  resourceNameInfo: () => `${config.resourceNamePlural}__info`,
  resourceNameCreate: () => `${config.resourceNamePlural}__create`,
  resourceNameModify: () => `${config.resourceNamePlural}__modify`,
  resourceNameRemove: () => `${config.resourceNamePlural}__remove`,
  resourceLogsInfo: () => `${config.resourceNameSingular}_logs__info`,
  resourceLogsFirst: () => `${config.resourceNameSingular}_logs__first`,
  resourceLogsLast: () => `${config.resourceNameSingular}_logs__last`,
  resourceLogsPrev: () => `${config.resourceNameSingular}_logs__prev`,
  resourceLogsNext: () => `${config.resourceNameSingular}_logs__next`,
  resourcePageFirst: () => `${config.resourceNamePlural}__first`,
  resourcePageLast: () => `${config.resourceNamePlural}__last`,
  resourcePagePrev: () => `${config.resourceNamePlural}__prev`,
  resourcePageNext: () => `${config.resourceNamePlural}__next`
}

module.exports = {
  make: make,
  initialize: initialize,
  configured: configured,
  paginate: paginate,
  raiseEvents: raiseEvents,
  retrieve: retrieve,
  store: store
}
