'use strict'

const AWS = require('aws-sdk')
const s3 = new AWS.S3({apiVersion: '2006-03-01', region: 'us-west-2'})
const meta = require('meta-ngin')
const controllers = require('./config')
const config = controllers.retrieve
const changeCase = require('change-case')
const wso2Request = require('byu-wso2-request').request

/**
 * Gets the JSON data from an S3 Bucket and returns only the JSON body. If getLogs is true, then it will get the data
 * from the logs file instead of the default storage file.
 * @param getLogs - boolean representing that the log file will or will not be retrieved.
 * @returns A promise that resolves when the data is returned from s3 and hands back the JSON body of the s3 bucket.
 */
exports.getObjects = getLogs => {
  let params = {
    Bucket: config.bucketName(),
    Key: (getLogs ? config.logFile() : config.storageFile())
  }
  return s3.getObject(params).promise()
    .then(data => JSON.parse(data.Body)[(getLogs ? `${config.resourceNameSingular()}_logs` : `${config.resourceNamePlural()}`)])
}

/**
 * Searches the s3 bucket data and returns code_sets that match the code and the sort.
 * @param data is an Array of code sets returned by the getObjects function.
 * @param code is the code_set(s) that are to be returned from the query.
 * @param sort is the sort group of which the code set must be a part in order to match the query.
 * @returns an Array of the code(s) (that are a part of the sort group if sort is specified) that match the query parameters.
 */
exports.getObject = (data, code, sort) => {
  let new_data = []

  if (sort) {
    data.forEach((code_set, index, data_array) => {
      if (code_set[config.sortNameSingular()] === sort &&
        code_set[config.resourceNameSingular()] === code) { new_data.push(data_array[index]) }
    })
  } else {
    data.forEach((code_set, index, data_array) => {
      if (code_set[config.resourceNameSingular()] === code) new_data.push(data_array[index])
    })
  }
  if (new_data.length === 0) throw new Error(`Resource not found`)
  return new_data
}

exports.putObject = (data, modifyLogs) => {
  let params = {
    Bucket: config.bucketName(),
    Key: (modifyLogs ? config.logFile() : config.storageFile()),
    Body: JSON.stringify({[(modifyLogs ? `${config.resourceNameSingular()}_logs` : config.resourceNamePlural())]: data})
  }

  return s3.putObject(params).promise()
    .then(() => data)
}

exports.log = (req, input, log_action) => {
  if (!log_action) throw new Error('Missing log_action parameter')
  if (log_action !== 'modify' && log_action !== 'create' && log_action !== 'remove') { throw new Error(`${log_action} is an invalid log action`) }

  exports.getObjects(true)
    .then(logs => {
      logs.push(Object.assign({
        log_action: `${config.resourceNameSingular()}__${log_action}`,
        updated_by_id: req.verifiedJWTs.current.byu.client.byuId,
        date_time_updated: new Date().toISOString()
      }, input))
      return logs
    })
    .then(new_logs => exports.putObject(new_logs, true))
}

exports.removeObject = (data, code, sort) => {
  let removed = []
  data.forEach((code_set, index, array) => {
    if (sort) {
      if (code_set[config.sortNameSingular()] === sort &&
        code_set[config.resourceNameSingular()] === code) { removed = array.splice(index, 1) } else {
        if (code_set[config.resourceNameSingular()] === code) { removed = array.splice(index, 1) }
      }
    }
  })
  if (removed.length <= 0) throw new Error(`Resource not found`)
  return data
}

/**
 * Calculates all pagination starting locations for first, last, prev, next
 * @param data - expects an array of the items that you want to paginate
 * @param page_start - expects a number for the start location (not by index)
 * @param page_size - expects a number for how many items to be returned
 * @returns {*}
 */
const paginate = (data, page_start, page_size) => {
  const collection_size = data.length
  page_start = parseInt(page_start)
  page_size = parseInt(page_size)

  const start = collection_size === 0 ? null : (page_start || 1) <= collection_size ? (page_start || 1) : 1
  let size = collection_size === 0 ? null : (page_size || config.pageDefaultSize()) > collection_size - start ? collection_size - start + 1 : (page_size || config.pageDefaultSize())
  if (size > config.pageMaxSize()) size = config.pageMaxSize()
  const last = collection_size === 0 ? null : collection_size - size + 1
  const next = collection_size === 0 ? null : start === last ? null : start + size
  const prev = collection_size === 0 ? null : start === 1 ? null : (start - size) > 0 ? start - size : 1
  const end = collection_size === 0 ? null : start + size - 1
  const result = collection_size > 0 ? data.splice(start - 1, size) : []

  return {start, end, size, next, prev, last, collection_size, result}
}

/**
 * formats the links for a collection of data
 * @param query expects a query object (like the one returned from paginate)
 * @returns {{}}
 */
let formatCollectionLinks = (query, isLog) => {
  let links = {}
  if (!isLog) {
    links[config.resourcePageFirst()] = {
      'rel': `${config.resourcePageFirst()}`,
      'href': `https://api.byu.edu/domains/identity/${config.resourceNamePlural()}/v1${query.start ? '?page_start=1&page_size=' + query.size : ''}`,
      'method': 'GET'
    }
    if (query.last) {
      links[config.resourcePageLast()] = {
        'rel': `${config.resourcePageLast()}`,
        'href': `https://api.byu.edu/domains/identity/${config.resourceNamePlural()}/v1?page_start=${query.last}&page_size=${query.size}`,
        'method': 'GET'
      }
    }
    if (query.next) {
      links[config.resourcePageNext()] = {
        'rel': `${config.resourcePageNext()}`,
        'href': `https://api.byu.edu/domains/identity/${config.resourceNamePlural()}/v1?page_start=${query.next}&page_size=${query.size}`,
        'method': 'GET'
      }
    }
    if (query.prev) {
      links[config.resourcePagePrev()] = {
        'rel': `${config.resourcePagePrev()}`,
        'href': `https://api.byu.edu/domains/identity/${config.resourceNamePlural()}/v1?page_start=${query.prev}&page_size=${query.size}`,
        'method': 'GET'
      }
    }
  } else {
    links[config.resourceLogsFirst()] = {
      'rel': `${config.resourceLogsFirst()}`,
      'href': `https://api.byu.edu/domains/identity/${config.resourceNamePlural()}/v1${query.start ? '?page_start=1&page_size=' + query.size : ''}`,
      'method': 'GET'
    }
    if (query.last) {
      links[config.resourceLogsLast()] = {
        'rel': `${config.resourceLogsLast()}`,
        'href': `https://api.byu.edu/domains/identity/${config.resourceNamePlural()}/v1?page_start=${query.last}&page_size=${query.size}`,
        'method': 'GET'
      }
    }
    if (query.next) {
      links[config.resourceLogsNext()] = {
        'rel': `${config.resourceLogsNext()}`,
        'href': `https://api.byu.edu/domains/identity/${config.resourceNamePlural()}/v1?page_start=${query.next}&page_size=${query.size}`,
        'method': 'GET'
      }
    }
    if (query.prev) {
      links[config.resourceLogsPrev()] = {
        'rel': `${config.resourceLogsPrev()}`,
        'href': `https://api.byu.edu/domains/identity/${config.resourceNamePlural()}/v1?page_start=${query.prev}&page_size=${query.size}`,
        'method': 'GET'
      }
    }
  }
  return links
}

exports.formatRes = (data, code, isLog) => {
  if (!isLog) {
    data = Object.assign({
      'links': {
        [`${config.resourceNameInfo()}`]: {
          'rel': 'self',
          'href': `https://api.byu.edu/domains/identity/${[config.resourceNamePlural()]}/v1/${(config.sortNameSingular() ? `${data[config.sortNameSingular()]},` : ``)}${data[config.resourceNameSingular()]}`,
          'method': 'GET'
        },
        [`${config.resourceNameModify()}`]: {
          'rel': config.resourceNameModify(),
          'href': `https://api.byu.edu/domains/identity/${[config.resourceNamePlural()]}/v1/${(config.sortNameSingular() ? `${data[config.sortNameSingular()]},` : ``)}${data[config.resourceNameSingular()]}`,
          'method': 'PUT'
        },
        [`${config.resourceNameRemove()}`]: {
          'rel': config.resourceNameRemove(),
          'href': `https://api.byu.edu/domains/identity/${[config.resourceNamePlural()]}/v1/${(config.sortNameSingular() ? `${data[config.sortNameSingular()]},` : ``)}${data[config.resourceNameSingular()]}`,
          'method': 'DELETE'
        }
      },
      'metadata': meta((code || 200), null)
    }, data)
  }
  return data
}

exports.formatCollectionRes = (data, state, isLog, query) => {
  let out
  if (!isLog) {
    out = {
      'links': {
        [`${config.resourceNameInfo()}`]: {
          'rel': 'self',
          'href': `https://api.byu.edu/domains/identity/${config.resourceNamePlural()}/v1`,
          'method': 'GET'
        },
        [`${config.resourceNameCreate()}`]: {
          'rel': config.resourceNameCreate(),
          'href': `https://api.byu.edu/domains/identity/${config.resourceNamePlural()}/v1`,
          'method': 'POST'
        }
      },
      'metadata': {
        'validation_response': {
          'code': state ? state.statusCode : 200,
          'message': state ? state.message : 'Successful'
        }
      },
      'values': []
    }
    data.forEach(code_set => {
      let query_value_found = true
      let query_keys = Object.keys(query)
      for (let i = 0; i < query_keys.length; i++) {
        if (query_keys[i] !== 'page_start' && query_keys[i] !== 'page_size') {
          if (code_set[query_keys[i]] !== query[query_keys[i]]) query_value_found = false
        }
      }
      if (query_value_found) out.values.push(exports.formatRes(code_set, 200, (isLog || false)))
    })
  } else {
    out = {
      'links': {
        [`${config.resourceLogsInfo()}`]: {
          'rel': 'self',
          'href': `https://api.byu.edu/domains/identity/${config.resourceNamePlural()}/v1/${(config.sortNameSingular() ? `${data[0][config.sortNameSingular()]},` : ``)}${data[0][config.resourceNameSingular()]}/logs`,
          'method': 'GET'
        }
      },
      'metadata': {
        'validation_response': {
          'code': state ? state.statusCode : 200,
          'message': state ? state.message : 'Successful'
        }
      },
      'values': []
    }
    data.forEach(code_set => {
      out.values.push(exports.formatRes(code_set, 200, (isLog || false)))
    })
  }

  if (controllers.paginate()) {
    let page_info = paginate(out.values, query.page_start, query.page_size)
    out.links = Object.assign(out.links, formatCollectionLinks(page_info, isLog))
    out.metadata = Object.assign(out.metadata, {
      collection_size: page_info.collection_size,
      page_start: page_info.start,
      page_end: page_info.end,
      page_size: page_info.size,
      default_page_size: config.pageDefaultSize(),
      max_page_size: config.pageMaxSize()
    })
    out.values = page_info.result
  }
  return out
}

exports.raiseEvent = log_entry => {
  return wso2Request({
    url: 'https://api.byu.edu/domains/eventhub/v1/events',
    method: 'POST',
    json: true,
    body: {
      events: {
        event: [
          {
            event_header: {
              domain: 'edu.byu',
              entity: 'BYU-IAM',
              event_type: `${changeCase.titleCase(config.resourceNameSingular())} Updated`
            },
            filters: '',
            event_body: log_entry
          }
        ]
      }
    }
  })
}
